@@ -1,14 +1,22 @@
 import asyncio
 import contextlib
+import inspect
 import warnings
-from collections.abc import Callable
-from typing import Any, Awaitable, Callable, Dict, Generator, Optional, Union
+from typing import (
+    Any,
+    Awaitable,
+    Callable,
+    Dict,
+    Iterator,
+    Optional,
+    Protocol,
+    Type,
+    Union,
+    overload,
+)
 
 import pytest
 
-from aiohttp.helpers import PY_37, isasyncgenfunction
-from aiohttp.web import Application
-
 from .test_utils import (
     BaseTestServer,
     RawTestServer,
@@ -19,18 +27,44 @@ from .test_utils import (
     teardown_test_loop,
     unused_port as _unused_port,
 )
+from .web import Application, BaseRequest, Request
+from .web_protocol import _RequestHandler
 
 try:
     import uvloop
 except ImportError:  # pragma: no cover
-    uvloop = None
+    uvloop = None  # type: ignore[assignment]
 
-try:
-    import tokio
-except ImportError:  # pragma: no cover
-    tokio = None
 
-AiohttpClient = Callable[[Union[Application, BaseTestServer]], Awaitable[TestClient]]
+class AiohttpClient(Protocol):
+    @overload
+    async def __call__(
+        self,
+        __param: Application,
+        *,
+        server_kwargs: Optional[Dict[str, Any]] = None,
+        **kwargs: Any,
+    ) -> TestClient[Request, Application]: ...
+    @overload
+    async def __call__(
+        self,
+        __param: BaseTestServer,
+        *,
+        server_kwargs: Optional[Dict[str, Any]] = None,
+        **kwargs: Any,
+    ) -> TestClient[BaseRequest, None]: ...
+
+
+class AiohttpServer(Protocol):
+    def __call__(
+        self, app: Application, *, port: Optional[int] = None, **kwargs: Any
+    ) -> Awaitable[TestServer]: ...
+
+
+class AiohttpRawServer(Protocol):
+    def __call__(
+        self, handler: _RequestHandler, *, port: Optional[int] = None, **kwargs: Any
+    ) -> Awaitable[RawTestServer]: ...
 
 
 def pytest_addoption(parser):  # type: ignore[no-untyped-def]
@@ -44,7 +78,7 @@ def pytest_addoption(parser):  # type: ignore[no-untyped-def]
         "--aiohttp-loop",
         action="store",
         default="pyloop",
-        help="run tests with specific loop: pyloop, uvloop, tokio or all",
+        help="run tests with specific loop: pyloop, uvloop or all",
     )
     parser.addoption(
         "--aiohttp-enable-loop-debug",
@@ -61,7 +95,7 @@ def pytest_fixture_setup(fixturedef):  # type: ignore[no-untyped-def]
     """
     func = fixturedef.func
 
-    if isasyncgenfunction(func):
+    if inspect.isasyncgenfunction(func):
         # async generator fixture
         is_async_gen = True
     elif asyncio.iscoroutinefunction(func):
@@ -193,16 +227,14 @@ def pytest_generate_tests(metafunc):  # type: ignore[no-untyped-def]
         return
 
     loops = metafunc.config.option.aiohttp_loop
+    avail_factories: Dict[str, Type[asyncio.AbstractEventLoopPolicy]]
     avail_factories = {"pyloop": asyncio.DefaultEventLoopPolicy}
 
     if uvloop is not None:  # pragma: no cover
         avail_factories["uvloop"] = uvloop.EventLoopPolicy
 
-    if tokio is not None:  # pragma: no cover
-        avail_factories["tokio"] = tokio.EventLoopPolicy
-
     if loops == "all":
-        loops = "pyloop,uvloop?,tokio?"
+        loops = "pyloop,uvloop?"
 
     factories = {}  # type: ignore[var-annotated]
     for name in loops.split(","):
@@ -236,12 +268,8 @@ def loop(loop_factory, fast, loop_debug):  # type: ignore[no-untyped-def]
 
 @pytest.fixture
 def proactor_loop():  # type: ignore[no-untyped-def]
-    if not PY_37:
-        policy = asyncio.get_event_loop_policy()
-        policy._loop_factory = asyncio.ProactorEventLoop  # type: ignore[attr-defined]
-    else:
-        policy = asyncio.WindowsProactorEventLoopPolicy()  # type: ignore[attr-defined]
-        asyncio.set_event_loop_policy(policy)
+    policy = asyncio.WindowsProactorEventLoopPolicy()  # type: ignore[attr-defined]
+    asyncio.set_event_loop_policy(policy)
 
     with loop_context(policy.new_event_loop) as _loop:
         asyncio.set_event_loop(_loop)
@@ -249,7 +277,7 @@ def proactor_loop():  # type: ignore[no-untyped-def]
 
 
 @pytest.fixture
-def unused_port(aiohttp_unused_port):  # type: ignore[no-untyped-def] # pragma: no cover
+def unused_port(aiohttp_unused_port: Callable[[], int]) -> Callable[[], int]:
     warnings.warn(
         "Deprecated, use aiohttp_unused_port fixture instead",
         DeprecationWarning,
@@ -259,20 +287,22 @@ def unused_port(aiohttp_unused_port):  # type: ignore[no-untyped-def] # pragma:
 
 
 @pytest.fixture
-def aiohttp_unused_port():  # type: ignore[no-untyped-def]
+def aiohttp_unused_port() -> Callable[[], int]:
     """Return a port that is unused on the current host."""
     return _unused_port
 
 
 @pytest.fixture
-def aiohttp_server(loop):  # type: ignore[no-untyped-def]
+def aiohttp_server(loop: asyncio.AbstractEventLoop) -> Iterator[AiohttpServer]:
     """Factory to create a TestServer instance, given an app.
 
     aiohttp_server(app, **kwargs)
     """
     servers = []
 
-    async def go(app, *, port=None, **kwargs):  # type: ignore[no-untyped-def]
+    async def go(
+        app: Application, *, port: Optional[int] = None, **kwargs: Any
+    ) -> TestServer:
         server = TestServer(app, port=port)
         await server.start_server(loop=loop, **kwargs)
         servers.append(server)
@@ -298,14 +328,16 @@ def test_server(aiohttp_server):  # type: ignore[no-untyped-def]  # pragma: no c
 
 
 @pytest.fixture
-def aiohttp_raw_server(loop):  # type: ignore[no-untyped-def]
+def aiohttp_raw_server(loop: asyncio.AbstractEventLoop) -> Iterator[AiohttpRawServer]:
     """Factory to create a RawTestServer instance, given a web handler.
 
     aiohttp_raw_server(handler, **kwargs)
     """
     servers = []
 
-    async def go(handler, *, port=None, **kwargs):  # type: ignore[no-untyped-def]
+    async def go(
+        handler: _RequestHandler, *, port: Optional[int] = None, **kwargs: Any
+    ) -> RawTestServer:
         server = RawTestServer(handler, port=port)
         await server.start_server(loop=loop, **kwargs)
         servers.append(server)
@@ -333,9 +365,7 @@ def raw_test_server(  # type: ignore[no-untyped-def]  # pragma: no cover
 
 
 @pytest.fixture
-def aiohttp_client(
-    loop: asyncio.AbstractEventLoop,
-) -> Generator[AiohttpClient, None, None]:
+def aiohttp_client(loop: asyncio.AbstractEventLoop) -> Iterator[AiohttpClient]:
     """Factory to create a TestClient instance.
 
     aiohttp_client(app, **kwargs)
@@ -344,13 +374,28 @@ def aiohttp_client(
     """
     clients = []
 
+    @overload
+    async def go(
+        __param: Application,
+        *,
+        server_kwargs: Optional[Dict[str, Any]] = None,
+        **kwargs: Any,
+    ) -> TestClient[Request, Application]: ...
+
+    @overload
+    async def go(
+        __param: BaseTestServer,
+        *,
+        server_kwargs: Optional[Dict[str, Any]] = None,
+        **kwargs: Any,
+    ) -> TestClient[BaseRequest, None]: ...
+
     async def go(
         __param: Union[Application, BaseTestServer],
         *args: Any,
         server_kwargs: Optional[Dict[str, Any]] = None,
-        **kwargs: Any
-    ) -> TestClient:
-
+        **kwargs: Any,
+    ) -> TestClient[Any, Any]:
         if isinstance(__param, Callable) and not isinstance(  # type: ignore[arg-type]
             __param, (Application, BaseTestServer)
         ):
